#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define TRUE 1
#define FALSE 0

typedef struct dposn {
  double x;
  double y;
} dposn;

dposn* make_dposn(double x, double y) {
  dposn* dp;
  dp = (dposn*)malloc(sizeof(dposn));
  dp->x = x;
  dp->y = y;
  return dp;
}

void free_dposn(dposn *dp) {
  free(dp);
}

/* ------------------------------------ */

enum {
  circle_type,
  rectangle_type
};

typedef struct shape {
  int type;
} shape;

typedef struct circle {
  shape s;
  double radius;
  dposn* center;
} circle;

typedef struct rectangle {
  shape s;
  dposn* topleft;
  double width;
  double height;
} rectangle;

circle* make_circle(double radius, dposn* center) {
  circle* c;
  c = (circle*)malloc(sizeof(circle));
  c->s.type = circle_type;
  c->radius = radius;
  c->center = make_dposn(center->x, center->y);
  return c;
}

void free_circle(circle* c) {
  free_dposn(c->center);
  free(c);
}

rectangle* make_rectangle(dposn* topleft, double width, double height) {
  rectangle* r;
  r = (rectangle*)malloc(sizeof(rectangle));
  r->s.type = rectangle_type;
  r->topleft = make_dposn(topleft->x, topleft->y);
  r->width = width;
  r->height = height;
  return r;
}

void free_rectangle(rectangle* r) {
  free_dposn(r->topleft);
  free(r);
}

/* ------------------------------------ */

double sqr(double v) { return v * v; }

int pt_in_circle(dposn *dp, circle* c) {
  return (sqrt(sqr(c->center->x - dp->x)
               + sqr(c->center->y - dp->y))
          <= c->radius);          
}

int pt_in_rectangle(dposn *dp, rectangle* r) {
  return ((r->topleft->x <= dp->x)
          && (r->topleft->x + r->width >= dp->x)
          && (r->topleft->y <= dp->y)
          && (r->topleft->y + r->height >= dp->y));
}

int pt_in_shape(dposn *dp, shape* s) {
  switch (s->type) {
  case circle_type:
    return pt_in_circle(dp, (circle*)s);
  case rectangle_type:
    return pt_in_rectangle(dp, (rectangle*)s);
  default:
    return 0;
  }
}

/* ------------------------------------ */

void failed(int a, int b) {
  printf("test failed; expected %d, got %d\n", b, a);
  exit(1);
}

void check_int(int a, int b) {
  if (a != b) failed(a, b);
}

int main(int argc, char** argv)
{
  dposn* origin = make_dposn(0, 0);
  circle *c1 = make_circle(3.5, origin);
  rectangle *r1 = make_rectangle(origin, 3.2, 6.7);
  dposn* oneone = make_dposn(1, 1);
  dposn* onenone = make_dposn(1, -1);
  dposn* onefive = make_dposn(1, 5);

  check_int(pt_in_shape(oneone, (shape*)c1), TRUE);
  check_int(pt_in_shape(oneone, (shape*)r1), TRUE);

  check_int(pt_in_shape(onenone, (shape*)c1), TRUE);
  check_int(pt_in_shape(onenone, (shape*)r1), FALSE);

  check_int(pt_in_shape(onefive, (shape*)c1), FALSE);
  check_int(pt_in_shape(onefive, (shape*)r1), TRUE);

  free_circle(c1);
  free_rectangle(r1);
  free_dposn(origin);
  free_dposn(oneone);
  free_dposn(onenone);
  free_dposn(onefive);

  printf("ok\n");

  return 0;
}
